// << ------------------------------------ >>
// << SLADE - SlayeR's LeetAss Doom Editor >>
// << By Simon Judd, 2004-05               >>
// << ------------------------------------ >>
// << texture_list.cpp: TextureList class  >>
// << functions                            >>
// << ------------------------------------ >>

// INCLUDES ------------------------------ >>
#include "main.h"

// VARIABLES ----------------------------- >>
static const WORD valid_dimensions[] = { 2, 4, 8, 16, 32, 64, 128, 256, 512, 1024, 2048, 4096 };

// EXTERNAL VARIABLES -------------------- >>
extern bool mix_tex;

// is_valid_dimension: Checks if a value is a power of 2
// -------------------------------------------------- >>
bool is_valid_dimension(WORD val)
{
	for (BYTE d = 0; d < 12; d++)
	{
		if (val == valid_dimensions[d])
			return true;
	}

	return false;
}

// up_dimension: Increases a dimension to be a power of two
// ----------------------------------------------------- >>
WORD up_dimension(WORD val)
{
	WORD ret = val;

	while (1)
	{
		ret++;

		for (BYTE d = 0; d < 12; d++)
		{
			if (ret == valid_dimensions[d])
				return ret;
		}
	}

	return 0;
}

// TextureList::add_tex: Adds a blank texture to the list
// --------------------------------------------------- >>
void TextureList::add_tex(string name, WORD width, WORD height, bool available, bool flat)
{
	if (tex_exists(name, available, flat))
		delete_tex(name);

	DWORD n = n_textures;
	n_textures++;
	textures = (texture_t **)realloc(textures, n_textures * sizeof(texture_t *));
	textures[n] = new texture_t;

	textures[n]->width = width;
	textures[n]->height = height;
	textures[n]->rwidth = width;
	textures[n]->rheight = height;
	textures[n]->name = name;
	textures[n]->available = available;
	textures[n]->flat = flat;
	textures[n]->data = NULL;
}

void TextureList::delete_tex(string name)
{
	DWORD index = get_tex_index(name);
	free(textures[index]);

	for (DWORD t = index; t < n_textures - 1; t++)
		textures[t] = textures[t + 1];

	n_textures--;
	textures = (texture_t **)realloc(textures, n_textures * sizeof(texture_t *));
}

// TextureList::init_tex_data: Inits the texture's temp data
// ------------------------------------------------------ >>
void TextureList::init_tex_data()
{
	DWORD n = n_textures - 1;

	if (!is_valid_dimension(textures[n]->width))
	{
		textures[n]->rwidth = up_dimension(textures[n]->width);
		textures[n]->tex_x = (float)textures[n]->width / (float)textures[n]->rwidth;
	}

	if (!is_valid_dimension(textures[n]->height))
	{
		textures[n]->rheight = up_dimension(textures[n]->height);
		textures[n]->tex_y = (float)textures[n]->height / (float)textures[n]->rheight;
	}

	textures[n]->set_data_size(textures[n]->rwidth, textures[n]->rheight);
	textures[n]->clear_data();
}

// TextureList::create_tex: Creates the last texture from it's temp data
// ------------------------------------------------------------------ >>
void TextureList::create_tex(int filter, BYTE flags)
{
	DWORD n = n_textures - 1;

	// Generate gl tex
	if (filter == 1)
	{
		glGenTextures(1, &textures[n]->tex_id);
		glBindTexture(GL_TEXTURE_2D, textures[n]->tex_id);
		glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_NEAREST);
		glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_NEAREST);
		glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, textures[n]->rwidth, textures[n]->rheight, 0, GL_RGBA, IL_UNSIGNED_BYTE, textures[n]->data);
	}

	if (filter == 2)
	{
		glGenTextures(1, &textures[n]->tex_id);
		glBindTexture(GL_TEXTURE_2D, textures[n]->tex_id);
		glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
		glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
		glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, textures[n]->rwidth, textures[n]->rheight, 0, GL_RGBA, IL_UNSIGNED_BYTE, textures[n]->data);
	}

	if (filter == 3)
	{
		glGenTextures(1, &textures[n]->tex_id);
		glBindTexture(GL_TEXTURE_2D, textures[n]->tex_id);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR_MIPMAP_LINEAR);
		gluBuild2DMipmaps(GL_TEXTURE_2D, GL_RGBA, textures[n]->rwidth, textures[n]->rheight, GL_RGBA, GL_UNSIGNED_BYTE, textures[n]->data);
	}

	free(textures[n]->data);
	textures[n]->data = NULL;
}

// TextureList::add_file: Loads a texture from a file (can be any DevIL supported format)
// ----------------------------------------------------------------------------------- >>
bool TextureList::add_file(string name, string filename, int filter, BYTE flags)
{
	GLuint id;
	int width = 0;
	int height = 0;

	ilGenImages(1, &id);
	ilBindImage(id);

	char path[128] = "";
	sprintf(path, "%s", filename.c_str());

	if (!ilLoadImage(path))
	{
		printf("Texture %s could not be loaded!\n", name);
		return false;
	}

	DWORD n = n_textures;
	n_textures++;
	textures = (texture_t **)realloc(textures, n_textures * sizeof(texture_t *));
	textures[n] = new texture_t;

	textures[n]->width = ilGetInteger(IL_IMAGE_WIDTH);
	textures[n]->height = ilGetInteger(IL_IMAGE_HEIGHT);
	textures[n]->name = name;
	textures[n]->tex_id = id;
	textures[n]->data = NULL;
	textures[n]->available = false;
	textures[n]->flat = false;

	// Check image dimensions
	if (!is_valid_dimension(ilGetInteger(IL_IMAGE_WIDTH)))
	{
		// Resize X dimension
		width = up_dimension(ilGetInteger(IL_IMAGE_WIDTH));
		textures[n]->tex_x = (float)ilGetInteger(IL_IMAGE_WIDTH) / (float)width;
		iluImageParameter(ILU_PLACEMENT, ILU_UPPER_LEFT);
		iluEnlargeCanvas(width, ilGetInteger(IL_IMAGE_HEIGHT), ilGetInteger(IL_IMAGE_DEPTH));
	}
	else
		width = ilGetInteger(IL_IMAGE_WIDTH);

	if (!is_valid_dimension(ilGetInteger(IL_IMAGE_HEIGHT)))
	{
		// Resize Y dimension
		height = up_dimension(ilGetInteger(IL_IMAGE_HEIGHT));
		textures[n]->tex_y = (float)ilGetInteger(IL_IMAGE_HEIGHT) / (float)height;
		iluImageParameter(ILU_PLACEMENT, ILU_UPPER_LEFT);
		iluEnlargeCanvas(ilGetInteger(IL_IMAGE_WIDTH), height, ilGetInteger(IL_IMAGE_DEPTH));
	}
	else
		height = ilGetInteger(IL_IMAGE_HEIGHT);

	// Generate gl tex
	if (filter == 1)
	{
		glGenTextures(1, &id);
		glBindTexture(GL_TEXTURE_2D, id);
		glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_NEAREST);
		glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_NEAREST);
		glTexImage2D(GL_TEXTURE_2D, 0, ilGetInteger(IL_IMAGE_BPP), ilGetInteger(IL_IMAGE_WIDTH),
			ilGetInteger(IL_IMAGE_HEIGHT), 0, ilGetInteger(IL_IMAGE_FORMAT), IL_UNSIGNED_BYTE, ilGetData());
	}

	if (filter == 2)
	{
		glGenTextures(1, &id);
		glBindTexture(GL_TEXTURE_2D, id);
		glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
		glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
		glTexImage2D(GL_TEXTURE_2D, 0, ilGetInteger(IL_IMAGE_BPP), ilGetInteger(IL_IMAGE_WIDTH),
			ilGetInteger(IL_IMAGE_HEIGHT), 0, ilGetInteger(IL_IMAGE_FORMAT), IL_UNSIGNED_BYTE, ilGetData());
	}

	if (filter == 3)
	{
		glGenTextures(1, &id);
		glBindTexture(GL_TEXTURE_2D, id);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR_MIPMAP_LINEAR);
		gluBuild2DMipmaps(GL_TEXTURE_2D, ilGetInteger(IL_IMAGE_BPP), ilGetInteger(IL_IMAGE_WIDTH), 
			ilGetInteger(IL_IMAGE_HEIGHT), ilGetInteger(IL_IMAGE_FORMAT), GL_UNSIGNED_BYTE, ilGetData());
	}

	//printf("loaded ok\n");
	return true;
}

GLuint TextureList::gl_tex(string name)
{
	for (DWORD t = 0; t < n_textures; t++)
	{
		if (str_upper(textures[t]->name) == str_upper(name))
			return textures[t]->tex_id;
	}

	return gl_tex("_notex");
}

DWORD TextureList::get_tex_index(string name)
{
	for (DWORD t = 0; t < n_textures; t++)
	{
		if (str_upper(textures[t]->name) == str_upper(name))
			return t;
	}

	return get_tex_index("_notex");
}

texture_t* TextureList::get_tex(string name)
{
	for (DWORD t = 0; t < n_textures; t++)
	{
		if (str_upper(textures[t]->name) == str_upper(name))
			return textures[t];
	}

	return get_tex("_notex");
}

texture_t* TextureList::get_tex(string name, bool available, bool flat)
{
	for (DWORD t = 0; t < n_textures; t++)
	{
		if (str_upper(textures[t]->name) == str_upper(name))
		{
			if (mix_tex)
			{
				if (textures[t]->available == available)
					return textures[t];
			}
			else
			{
				if (textures[t]->available == available && textures[t]->flat == flat)
					return textures[t];
			}
		}
	}

	return get_tex("_notex");
}

bool TextureList::tex_exists(string name, bool available, bool flat)
{
	for (DWORD t = 0; t < n_textures; t++)
	{
		if (str_upper(textures[t]->name) == str_upper(name))
		{
			if (mix_tex)
			{
				if (textures[t]->available == available)
					return true;
			}
			else
			{
				if (textures[t]->available == available && (!mix_tex && textures[t]->flat == flat))
					return true;
			}
		}
	}

	return false;
}

string TextureList::get_name(GLuint id)
{
	for (DWORD t = 0; t < n_textures; t++)
	{
		if (textures[t]->tex_id == id)
			return textures[t]->name;
	}

	return "_notex";
}

texture_t* TextureList::get_last()
{
	return textures[n_textures - 1];
}